import unittest
from krux.para import *
from krux.para.exceptions import *


class ParentPara(Para):
    a = F(name='A', default=100)
    b = F(name='B', required=True, vmin=200, vmax=400)


class ParentMixin:
    a = F(name='A')
    b = F(name='B')


class ChildPara(ParentPara):
    c = F(name='C')


class ChildParaByMixin(ParentMixin, Para):
    c = F(name='C')


class WithCompute(Para):
    a = F('A', default=100)
    b = F('B', default=lambda para: para.a * 2, vmin=200, vmax=400)


class MyTestCase(unittest.TestCase):
    # test meta class
    def test_items_removed_from_class_definition(self):
        self.assertIsNone(getattr(ParentPara, 'a', None))
        self.assertIsNone(getattr(ParentPara, 'b', None))

    def test_subclass(self):
        self.assertListEqual(list(ChildPara._fields.keys()), ['a', 'b', 'c'])

    def test_mixin(self):
        self.assertListEqual(list(ChildParaByMixin._fields.keys()), ['a', 'b', 'c'])

    # test creation, validation and conversion
    def test_init_and_validation(self):
        para1 = ParentPara(a=200, b=300)
        para2 = ParentPara(b=300)
        with self.assertRaises(MissingRequiredParaField):
            para3 = ParentPara()

        with self.assertRaises(ParaFieldValueOutOfBound):
            para4 = ParentPara(b=100)

        with self.assertRaises(ParaFieldValueOutOfBound):
            para4 = ParentPara(b=500)

    def test_from_dict_and_to_dict(self):
        d1 = {"a": 200, "b": 300, "c": 123}
        para1 = ChildPara.from_dict(d1)
        self.assertDictEqual(para1.to_dict(), d1)

    def test_copy_and_vary(self):
        raw = {"b": 300, "c": 123}
        para1 = ChildPara.from_dict(raw)
        para2 = para1.copy()
        self.assertDictEqual(para1.to_dict(), para2.to_dict())

        para3 = para1.vary(a=200, c=256)
        self.assertDictEqual(para3.to_dict(), {"a": 200, "b": 300, "c": 256})

    def test_with_compute(self):
        para1 = WithCompute()
        para2 = WithCompute(a=120)
        para3 = WithCompute(b=300)

        self.assertDictEqual(para1.to_dict(), {"a": 100, "b": 200})
        self.assertDictEqual(para2.to_dict(), {"a": 120, "b": 240})
        self.assertDictEqual(para3.to_dict(), {"a": 100, "b": 300})

        with self.assertRaises(ParaFieldValueOutOfBound):
            para_over = WithCompute(a=250)

    def test_getattr(self):
        para = WithCompute(a=120)
        self.assertEqual(para.a, 120)
        self.assertEqual(para.b, 240)

    def test_dict_compatible(self):
        para = WithCompute(a=120)
        self.assertEqual(para['a'], 120)
        self.assertEqual(para['b'], 240)

        self.assertSetEqual(set(para.keys()), {"a", "b"})
        self.assertSetEqual(set(para.values()), {120, 240})
        self.assertListEqual([(key, value) for key, value in para.items()], [("a", 120), ("b", 240)])

    def test_invalid_field(self):
        d = {"a": 200, "b": 300, "c": 123}
        para = ChildPara.from_dict(d)

        with self.assertRaises(InvalidParaField):
            para.x

        with self.assertRaises(InvalidParaField):
            para['x']


if __name__ == '__main__':
    unittest.main()



__all__ = ['none_min', 'none_max', 'json_safe']

from datetime import date


def none_min(a, b):
    if a is None:
        return b
    elif b is None:
        return a
    return min(a, b)


def none_max(a, b):
    if a is None:
        return b
    elif b is None:
        return a
    return max(a, b)


def json_safe(data, time_format='%Y-%m-%d %H:%M;%S'):
    import numpy as np

    if isinstance(data, dict):
        return {k: json_safe(v) for k, v in data.items()}
    elif isinstance(data, (tuple, list, np.ndarray)):
        return [json_safe(v) for v in data]
    else:
        try:
            if isinstance(data, date):
                return data.strftime(time_format)
            elif np.isscalar(data):
                return None if not np.isfinite(data) else data
    except Exception as e:
            pass

        return data




__all__ = ['batched', 'batch']


def batched(iterable, n):
    """
    Author: Eric Cotner
    Source: https://discuss.python.org/t/add-batching-function-to-itertools-module/19357
    Conforms to Python 3.12 itertools.batched
    Batches an iterable into lists of given maximum size, yielding them one by one.
    """
    batch = []
    for element in iterable:
        batch.append(element)
        if len(batch) >= n:
            yield batch
            batch = []
    if len(batch) > 0:
        yield batch


def batch(array, size=1):
    """
    Creates an array of elements split into groups the length of size. If array can't be split evenly, the final chunk will be the remaining elements.
    Conform to lodash.batch
    """
    return batched(array, size)



# -*- coding:utf-8 -*-

import six
import string
from random import choice
from hashlib import md5

__all__ = [
    'md5_uuid',
    'random_str',
    'gen_uuid4', 'gen_uuid8', 'gen_uuid10', 'gen_uuid12',
    'gen_uuid16', 'gen_uuid20', 'gen_uuid24', 'gen_uuid32',
    'gen_uuid48', 'gen_uuid64', 'gen_uuid96', 'gen_uuid128',
]

letters = string.letters if six.PY2 else string.ascii_letters
RANDOM_CHARS = letters + string.digits


def md5_uuid(s, length=8):
    if (six.PY2 and isinstance(s, unicode)) or (six.PY3 and isinstance(s, str)):
        s = s.encode('utf-8')
    m = md5()
    m.update(s)
    return m.hexdigest()[:length]


def random_str(length=16, chars=RANDOM_CHARS):
    """generates random string.
    """
    return ''.join([choice(chars) for i in range(length)])


def gen_uuid4():
    return random_str(4)


def gen_uuid6():
    return random_str(6)


def gen_uuid8():
    return random_str(8)


def gen_uuid10():
    return random_str(10)


def gen_uuid12():
    return random_str(12)


def gen_uuid16():
    return random_str(16)


def gen_uuid20():
    return random_str(20)


def gen_uuid24():
    return random_str(24)


def gen_uuid32():
    return random_str(32)


def gen_uuid48():
    return random_str(48)


def gen_uuid64():
    return random_str(64)


def gen_uuid96():
    return random_str(96)


def gen_uuid128():
    return random_str(128)



__all__ = [
    'Schema',
    'SchemaItem',
    'NullDefaultValue',
]

from .exceptions import *


class NullDefaultValue:
    pass


class SchemaItem:
    def __init__(self, default=NullDefaultValue, required=False, desc=''):
        self.default = default
        self.required = required
        self.desc = desc


class Schema:
    Item = SchemaItem

    def __init__(self, dic=None):
        if not dic:
            dic = {}
        self._dic = dic

        self._required_keys = { key for (key, item) in self._dic.items() if item.required }
        self._defaults = {key: item.default for (key, item) in self._dic.items() if item.default is not NullDefaultValue}

    def validate(self, data):
        if not self:
            return {}

        missing_keys = self._required_keys - set(self._defaults.keys()) - set(data.keys())
        if len(missing_keys) > 0:
            raise SchemaMissingRequiredKeys(missing_keys)

        cleaned = self._defaults.copy()
        for key, value in data.items():
            # TODO: add type checking and other validations
            if key in self._dic:
                cleaned[key] = value

        return cleaned

    def __bool__(self):
        return bool(self._dic)



import unittest
from krux.values import *
import json
import numpy as np


class TestValues(unittest.TestCase):
    def test_none_min_max(self):
        self.assertEqual(none_min(1, 2), 1)
        self.assertEqual(none_min(1, None), 1)
        self.assertEqual(none_min(None, 2), 2)
        self.assertEqual(none_min(None, None), None)

        self.assertEqual(none_max(1, 2), 2)
        self.assertEqual(none_max(1, None), 1)
        self.assertEqual(none_max(None, 2), 2)
        self.assertEqual(none_max(None, None), None)

    def test_json_safe(self):
        d = {"a": np.nan, "b": np.inf, "c": [1, np.nan], "d": {"e": 1, "f": np.inf, "g": np.array([np.nan, 2, np.inf])}, "s": "abc"}
        s = json_safe(d)
        print(s)
        j = json.dumps(s)
        print(j)


if __name__ == '__main__':
    unittest.main()



# yet another version of Schema
__all__ = ['ParaBase', 'Para', 'ParaField', 'F']

from collections import OrderedDict
from krux.types.null import Null
from .exceptions import *


class ParaBase(type):
    """Meta class for para"""
    def __new__(cls, name, bases, attrs, **kwargs):
        super_new = super().__new__

        # Ensure initialization is only performed for subclasses of Para
        # (Excluding Para class itself)
        parents = [b for b in bases if isinstance(b, ParaBase)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        _fields = OrderedDict()

        # Create the class
        base_attrs = OrderedDict()
        for base in bases:
            if base_fields := getattr(base, '_fields', None):
                _fields.update(base_fields)

            for key, value in base.__dict__.items():
                if isinstance(value, ParaField):
                    _fields[key] = value
                elif key not in ('_fields', '__module__'):
                    base_attrs[key] = value

        to_pop = []
        for key, value in attrs.items():
            if isinstance(value, ParaField):
                _fields[key] = value
                to_pop.append(key)

        for key in to_pop:
            attrs.pop(key)

        module = attrs.pop('__module__')

        new_attrs = {
            '__module__': module,
            '_fields': _fields,
            **base_attrs,
            **attrs,
        }
        new_bases = tuple()

        new_class = super_new(cls, name, new_bases, new_attrs)
        return new_class


class Para(metaclass=ParaBase):
    # _fields will be modified/replaced by metaclass
    _fields = OrderedDict()

    def __init__(self, **kwargs):
        self._data = {}

        for key, field in self._fields.items():
            if field.default is not Null:
                self._data[key] = field.default

        for key, value in kwargs.items():
            if key in self._fields:
                self._data[key] = value

        self._validate_values()

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def to_dict(self):
        return {field: self[field] for field in self._fields}

    def copy(self):
        return self.__class__.from_dict(self._data)

    def vary(self, **kwargs):
        return self.__class__.from_dict({**self._data, **kwargs})

    def update(self, key_or_dic, value=None):
        if isinstance(key_or_dic, dict):
            self._data.update(key_or_dic)
            self._validate_values(keys=set(key_or_dic.keys()))
        else:
            self._data[key_or_dic] = value
            self._validate_values(keys={key_or_dic})

    def _validate_values(self, keys=None):
        for key, field in self._fields.items():
            if keys is not None and key not in keys:
                continue

            if key not in self._data:
                if field.required:
                    raise MissingRequiredParaField(key)
            else:
                raw_value = self._data[key]
                if callable(raw_value):
                    raw_value = raw_value(self)
                field.validate(raw_value)

    def __len__(self):
        return len(self._fields)

    # dict compatible methods
    def keys(self):
        return self._fields.keys()

    def values(self):
        for field in self.keys():
            yield self[field]

    def items(self):
        for field in self.keys():
            yield field, self[field]

    def __getitem__(self, field):
        if field not in self._fields:
            raise InvalidParaField(f'Invalid para field {field} for {self}.')

        if field in self._data:
            value = self._data[field]
            if callable(value):
                return value(self)
            else:
                return value
        else:
            return None

    def __getattr__(self, field):
        return self[field]

    def __repr__(self):
        return f'<{self.__class__.__name__}>'

    def preview(self):
        pass


class ParaField:
    def __init__(self,
                 name='', desc='',
                 default=Null,
                 required=False,
                 vmin=None, vmax=None,
                 **kwargs):
        self.name = name
        self.default = default
        self.required = required
        self.desc = desc
        self.vmin = vmin
        self.vmax = vmax

    def validate(self, value):
        if self.vmin is not None and value < self.vmin:
            raise ParaFieldValueOutOfBound(f'{value} is too small for {self}, min value: {self.vmin}.')

        if self.vmax is not None and value > self.vmax:
            raise ParaFieldValueOutOfBound(f'{value} is too large for {self}, max value: {self.vmax}.')

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name}>'


def F(name='', desc='',
      default=Null, required=False,
      vmin=None, vmax=None,
      **kwargs):
    # This shortcut function will become a factory for creating ParaField objects.
    # ParaField may be extended to more subclasses
    return ParaField(name=name, desc=desc,
                     default=default, required=required,
                     vmin=vmin, vmax=vmax,
                     **kwargs)


# -*- coding:utf-8 -*-

import six


__all__ = ['parse_bool']


def parse_bool(s):
    if isinstance(s, six.string_types):
        s = s.upper()
        if s in ('FALSE', 'F', 'NO', 'N', '0', '0.0', ''):
            return False
        elif s in ('TRUE', 'T', 'YES', 'Y', '1', '1.0'):
            return True
    return bool(s)



# -*- coding:utf-8 -*-

import six

__all__ = ['pick', 'omit', 'defaults', 'deep_update']


def pick(d, keys, attr=False):
    if attr:
        result = {k: getattr(d, k) for k in keys if hasattr(d, k)}
    else:
        result = {k: d[k] for k in keys if k in d}
    return result


def omit(d, keys):
    result = d.copy()
    for k in keys:
        result.pop(k, None)
    return result


def defaults(d1, d2, inplace=True):
    result = d1 if inplace else d1.copy()
    for k, v in six.iteritems(d2):
        result.setdefault(k, v)
    return result


def deep_update(a, b):
    """deep version of dict.update()"""
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                deep_update(a[key], b[key])
            elif a[key] == b[key]:
                pass
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a



import unittest
from krux.schema import *
from krux.schema.exceptions import *


class TestSchema(unittest.TestCase):
    def test_required_keys_and_default(self):
        schema = Schema({
            'a': Schema.Item(required=True),
            'b': Schema.Item(),
            'c': Schema.Item(default=1, required=True)
        })

        with self.assertRaises(SchemaMissingRequiredKeys):
            schema.validate({"c": 1})

        self.assertDictEqual(schema.validate({"a": 1}), {"a": 1, "c": 1})

        self.assertDictEqual(schema.validate({
            "a": 1, "b": 2, "c": 3, "d": 4
        }), {"a": 1, "b": 2, "c": 3})



# -*- coding:utf-8 -*-


class Singleton(object):
    """A do-nothing class.
    From A. Martelli et al. Python Cookbook. (O'Reilly)
    Thanks to Juergen Hermann."""
    def __new__(cls, *args, **kwargs):
        if '_inst' not in vars(cls):
            cls._inst = super(Singleton, cls).__new__(cls, *args, **kwargs)
        return cls._inst



# -*- coding:utf-8 -*-

__all__ = ['NullClass', 'Null']

from .singleton import Singleton


class NullClass(Singleton):
    """A do-nothing class.
    From A. Martelli et al. Python Cookbook. (O'Reilly)
    Thanks to Dinu C. Gherman, Holger Krekel.
    """
    def __init__(self, *args, **kwargs): pass
    def __call__(self, *args, **kwargs): return self
    def __repr__(self): return "Null"
    def __nonzero__(self): return False
    def __getattr__(self, name): return self
    def __setattr__(self, name, value): return  self
    def __delattr__(self, name): return self
    def __len__(self): return 0
    def __iter__(self): return iter(())
    def __getitem__(self, i): return self
    def __delitem__(self, i): return self
    def __setitem__(self, i): return self


Null = NullClass()
